class DrukarniaAPIError(Exception):
    def __init__(self, message: str, status_code: int or str, request_type: str, request_url: str):
        """
        Custom exception class for handling errors in the Drukarnia API.

        Args:
            message (str): The error message.
            status_code (int or str): The status code of the API response.
            request_type (str): The type of the API request.
            request_url (str): The URL of the API request.
        """
        self.message = message
        self.status_code = status_code
        self.request_type = request_type
        self.request_url = request_url

    def __str__(self):
        """
        Returns a string representation of the exception.
        """
        text = f"""
        Request Type: {self.request_type}
        Response Status: {self.status_code}
        Request To: {self.request_url}
        Error Message: "{self.message}"
        """

        return text


class DrukarniaElementDataError(Exception):
    def __init__(self, attr_name: str, solution: str):
        """
        Custom exception class for handling errors related to missing attributes in Drukarnia element data.

        Args:
            attr_name (str): The name of the missing attribute.
            method_name (str): The name of the method where the attribute was expected.
            solution (str): Possible solution to resolve the error.
        """
        self.attr_name = attr_name
        self.solution = solution

    def __str__(self):
        """
        Returns a string representation of the exception.
        """
        text = f"""
        Attribute {self.attr_name} has not been found.
        Possible solution: {self.solution}
        """

        return text



from datetime import datetime
from warnings import warn
from aiohttp import ClientSession

from drukarnia_api.drukarnia_base import DrukarniaElement
from drukarnia_api.shortcuts import data2authors, data2articles, data2tags

from typing import TYPE_CHECKING, Tuple, Dict, List

if TYPE_CHECKING:  # always False, used for type hints
    from drukarnia_api.author import Author
    from drukarnia_api.tag import Tag


class Article(DrukarniaElement):
    def __init__(self, slug: str = None, article_id: str = None, *args, **kwargs):
        """
        Initializes an Article object with the given slug and article ID.
        """
        super().__init__(*args, **kwargs)

        self._update_data({'slug': slug, '_id': article_id})

    @DrukarniaElement._control_attr('article_id')
    @DrukarniaElement._is_authenticated
    async def post_comment(self, comment_text: str) -> str:
        """
        Posts a comment on the article and returns the ID of the posted comment.
        """

        posted_comment_id = await self.request('post', '/api/articles/{_id}/comments'.format(_id=self.article_id),
                                               data={'comment': comment_text}, output='read')
        return str(posted_comment_id)

    @DrukarniaElement._control_attr('article_id')
    @DrukarniaElement._is_authenticated
    async def reply2comment(self, comment_text: str, comment_id: str, root_comment: str,
                            root_comment_owner: str, reply2user: str) -> str:
        """
        Posts a reply to a comment and returns the ID of the new comment.
        """

        new_comment_id = await self.request('post',
                                            f'/api/articles/{self.article_id}/comments/{comment_id}/replies',
                                            data={
                                                  "comment": comment_text, "replyToComment": comment_id,
                                                  "rootComment": root_comment,
                                                  "rootCommentOwner": root_comment_owner, "replyToUser": reply2user
                                                },
                                            output='read'
                                            )

        return new_comment_id.decode('utf-8')

    @DrukarniaElement._control_attr('article_id')
    @DrukarniaElement._is_authenticated
    async def delete_comment(self, comment_id: str) -> None:
        """
        Deletes a comment from the article.
        """

        await self.request('delete', f'/api/articles/{self.article_id}/comments/{comment_id}')

    @DrukarniaElement._control_attr('article_id')
    @DrukarniaElement._is_authenticated
    async def like_comment(self, comment_id: str, unlike: bool = False) -> None:
        """
        Likes or unlikes a comment based on the 'delete' parameter.
        """

        if unlike:
            await self.request('delete', f'/api/articles/{self.article_id}/comments/{comment_id}/likes')

        else:
            await self.request('post', f'/api/articles/{self.article_id}/comments/{comment_id}/likes')

    @DrukarniaElement._control_attr('article_id')
    @DrukarniaElement._is_authenticated
    async def like_article(self, n_likes: int) -> None:
        """
        Likes the article with the specified number of likes.
        """

        if not (0 <= n_likes <= 10):
            raise ValueError('Number of likes must be greater or equal to zero and lower or equal to ten')

        await self.request('post', f'/api/articles/{self.article_id}/like', data={'likes': n_likes})

    @DrukarniaElement._control_attr('article_id')
    @DrukarniaElement._is_authenticated
    async def bookmark(self, section_id: str = '', unbookmark: bool = False) -> None:
        """
        Adds or removes the article from bookmarks based on the 'unbookmark' parameter.
        """

        if unbookmark:
            await self.request('delete', f'/api/articles/{self.article_id}/bookmarks')

        elif not section_id:
            raise ValueError('section_id must be passed for bookmarking')

        else:
            await self.request('post', '/api/articles/bookmarks',
                               data={"article": self.article_id, "list": section_id})

    @DrukarniaElement._control_attr('slug')
    async def collect_data(self, return_: bool = False) -> Dict or None:
        """
        Collects the article's data and updates the object's attributes.
        If 'return_' is True, returns the collected data.
        """

        data = await self.request('get', f'/api/articles/{self.slug}', output='json')

        self._update_data(data)

        if return_:
            return data

    @property
    async def owner(self) -> 'Author' or None:
        """
        Retrieves the owner of the article.
        """
        owner = self._access_data('owner', None)
        if owner is None:
            return None

        return await data2authors([owner], self.session)

    @property
    def comments(self) -> List:
        """
        Retrieves the comments of the article.
        """
        return self._access_data('comments', [])

    @property
    async def recommended_articles(self) -> Tuple['Article']:
        """
        Retrieves the recommended articles related to the article.
        """
        return await data2articles(self._access_data('recommendedArticles', []), self.session)

    @property
    def relationships(self) -> Dict:
        """
        Retrieves the relationships of the article.
        """
        return self._get_basetype_from_data('relationships', dict)

    @property
    def is_bookmarked(self) -> bool:
        """
        Checks if the article is bookmarked.
        """
        return self._get_basetype_from_data('isBookmarked', bool)

    @property
    def is_liked(self) -> bool:
        """
        Checks if the article is liked.
        """
        return self._get_basetype_from_data('isLiked', bool)

    @property
    def sensitive(self) -> bool:
        """
        Checks if the article is sensitive.
        """
        return self._get_basetype_from_data('sensitive', bool)

    @property
    def content(self) -> Dict:
        """
        Retrieves the content of the article.
        """
        return self._get_basetype_from_data('content', dict)

    @property
    async def author_articles(self) -> Tuple['Article']:
        """
        Retrieves the articles written by the author of the article.
        """
        return await data2articles(self._access_data('authorArticles', []), self.session)

    @property
    def thumb_picture(self) -> str:
        """
        Retrieves the thumbnail picture of the article.
        """
        return self._get_basetype_from_data('thumbPicture', str)

    @property
    def picture(self) -> str:
        """
        Retrieves the picture of the article.
        """
        return self._get_basetype_from_data('picture', str)

    @property
    def ads(self) -> str:
        """
        Retrieves the ads of the article.
        """
        return self._get_basetype_from_data('ads', str)

    @property
    def index(self) -> str:
        """
        Retrieves the index of the article.
        """
        return self._get_basetype_from_data('index', str)

    @property
    def created_at(self) -> datetime:
        """
        Retrieves the creation date of the article.
        """
        return self._get_datetime_from_author_data('createdAt')

    @property
    def read_time(self) -> float:
        """
        Retrieves the read time of the article.
        """
        return self._get_basetype_from_data('readTime', float)

    @property
    def number_of_like(self) -> int:
        """
        Retrieves the number of likes of the article.
        """
        return self._get_basetype_from_data('description', int)

    @property
    def number_of_comment(self) -> int:
        """
        Retrieves the number of comments of the article.
        """
        return self._get_basetype_from_data('commentNum', int)

    @property
    async def article_tags(self) -> Tuple['Tag']:
        """
        Retrieves the tags of the article.
        """

        tags = self._access_data('tags', [])

        if any(map(lambda el: isinstance(el, str), tags)):
            warn("Tag's data is incomplete, use collect_data method before in order to obtain whole data.")
            tags = map(lambda el: {'_id': el}, tags)

        return await data2tags(tags, self.session)

    @property
    async def main_article_tag(self) -> 'Tag' or None:
        """
        Retrieves the main tag of the article.
        """

        main_id = self._get_basetype_from_data('mainTagId', str)
        main_name = self._get_basetype_from_data('mainTag', str)
        main_slug = self._get_basetype_from_data('mainTagSlug', str)

        if not (main_id and main_name and main_slug):
            return None

        main_tag, = await data2tags([{'_id': main_id, 'name': main_name, 'slug': main_slug}], self.session)

        return main_tag

    @property
    def description(self) -> str:
        """
        Retrieves the description of the article.
        """
        return self._get_basetype_from_data('description', str)

    @property
    def seo_title(self) -> str:
        """
        Retrieves the SEO title of the article.
        """
        return self._get_basetype_from_data('seoTitle', str)

    @property
    def title(self) -> str:
        """
        Retrieves the title of the article.
        """
        return self._get_basetype_from_data('title', str)

    @property
    def article_id(self) -> str:
        """
        Retrieves the ID of the article.
        """
        return self._get_basetype_from_data('_id', str)

    @property
    def slug(self) -> str:
        """
        Retrieves the slug of the article.
        """
        return self._get_basetype_from_data('slug', str)

    @staticmethod
    async def from_records(session: ClientSession, new_data: dict) -> 'Article':
        """
        Creates an Article instance from records.
        """
        new_article = Article(session=session)
        new_article._update_data(new_data)

        return new_article

    def __hash__(self) -> int:
        """
        Returns the hash value of the Article object.
        """
        return hash(self.article_id or self.slug)



from typing import Any, Callable
from warnings import warn
from datetime import datetime
import inspect
from drukarnia_api.drukarnia_base.connection import Connection
from drukarnia_api.drukarnia_base.exceptions import DrukarniaElementDataError


class DrukarniaElement(Connection):
    """
    A class representing a printing element in a printing shop.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.all_collected_data = {}

    def _get_basetype_from_data(self, key: str, type_: Any = int, default: Any = 'auto'):
        """
        Get the value of the specified key from the author data dictionary and cast it to the specified type.

        Args:
            key (str): The key to access the value in the author data dictionary.
            type_ (Any): The type to cast the value to.
            default (Any): The default value if the key is not found. If set to 'auto', the default value will be
                           an instance of the specified type.

        Returns:
            Any: The value from the author data dictionary casted to the specified type or the default value.

        """

        if default == 'auto':
            default = type_()

        check = lambda el1: el1 is default if default is None else el1 == default

        n = self._access_data(key, default)
        return type_(n) if check(n) else n

    def _get_datetime_from_author_data(self, key: str):
        """
        Get a datetime object from the specified key in the author data dictionary.

        Args:
            key (str): The key to access the value in the author data dictionary.

        Returns:
            datetime: A datetime object parsed from the value in the author data dictionary.
        """
        date = self._access_data(key)

        if date:
            date = datetime.fromisoformat(date[:-1])

        return date

    def _update_data(self, new_data: dict):
        """
        Update the collected data with new data.

        Args:
            new_data (dict): The dictionary containing the new data to be added.
        """
        self.all_collected_data.update(new_data)

    def _access_data(self, key: str, default: Any = None):
        """
        Access a value from the collected data using the specified key.

        Args:
            key (str): The key to access the value in the collected data dictionary.
            default (Any): The default value to return if the key is not found.

        Returns:
            Any: The value from the collected data dictionary corresponding to the key, or the default value.
        """
        return self.all_collected_data.get(key, default)

    @staticmethod
    def _control_attr(attr: str, solution: str = 'call collect_data before') -> Callable:
        def decorator(func):
            async def wrapper(self_instance, *args, **kwargs):
                if getattr(self_instance, attr) is None:
                    raise DrukarniaElementDataError(attr, solution)

                return await func(self_instance, *args, **kwargs)

            return wrapper

        return decorator

    @staticmethod
    def _is_authenticated(func):
        """
        Decorator to check if the headers contain a Cookie.

        Args:
            func (function): The function to decorate.

        Returns:
            function: The decorated function.
        """

        async def wrapper(self_instance, *args, **kwargs):
            """
            Wrapper function that performs the authentication check before calling the decorated function.

            Args:
                self_instance (DrukarniaElement): The instance of the DrukarniaElement class.
                *args: Variable length argument list.
                **kwargs: Arbitrary keyword arguments.

            Returns:
                Any: The result of the decorated function.
            """
            if 'Cookie' not in self_instance.session.headers:
                warn("Cookie data was not identified, it may cause an error for this request")

            return await func(self_instance, *args, **kwargs)

        return wrapper



import asyncio
import json

from aiohttp import ClientSession, ClientResponse
from typing import Any, Callable, Dict, Generator, Tuple, List, Iterable
import inspect
from drukarnia_api.drukarnia_base.exceptions import DrukarniaAPIError


async def _from_response(response: ClientResponse, output: str or List[str] or None) -> Any:

    if response.status not in [200, 201]:
        data = await response.json()
        raise DrukarniaAPIError(data['message'], response.status,
                                response.request_info.method, str(response.request_info.url))

    if isinstance(output, str):
        data = await _from_response(response, [output])
        return data[0]

    if output is None:
        return []

    data = []
    for func_name in output:
        attr = getattr(response, func_name)

        if inspect.iscoroutinefunction(attr):
            data.append(await attr())

        elif callable(attr):
            data.append(attr())

        else:
            data.append(attr)

    return data


class Connection:
    base_url = 'https://drukarnia.com.ua'

    def __init__(self, session: ClientSession = None, headers: dict = None,
                 create_user_agent: bool = True, *args, **kwargs):

        # Save the aiohttp session
        if session:
            self.session = session
        else:
            headers_ = {}
            if headers:
                headers_ = headers

            if create_user_agent:
                from fake_useragent import UserAgent

                headers_['User-Agent'] = UserAgent().random

            self.session = ClientSession(base_url=self.base_url, headers=headers_, *args, **kwargs)

    def _update_headers(self, new_data, inplace: bool = True) -> dict:
        if inplace:
            self.session.headers.update(new_data)

        return dict(self.session.headers) | new_data

    async def request(self, method: str, url: str, output: str or list = None, **kwargs) -> Any:

        if (method in ['post', 'put', 'patch']) and isinstance(kwargs.get('data', {}), dict):
            # if data parameter will be passed as None, (e.g. not passed at all)
            # aiohttp will switch to aplication/octet media type
            # which is not supported by Drukarnia API

            kwargs['data'] = json.dumps(kwargs.get('data', {}))
            headers = self._update_headers({'Content-Type': 'application/json'}, inplace=False)

            async with self.session.request(method.upper(), url, headers=headers, **kwargs) as response:
                return await _from_response(response, output)

        async with self.session.request(method.upper(), url, **kwargs) as response:
            return await _from_response(response, output)

    async def request_pool(self, heuristics: Iterable[Dict[str, Any]]) -> Tuple:
        # Create tasks
        tasks = [self.request(**kwargs)
                 for kwargs in heuristics]

        # Get results
        return await asyncio.gather(*tasks)

    async def run_until_no_stop(self, request_synthesizer: Generator[Dict or List[Dict, str], None, None],
                                not_stop_until: Callable[[Any], bool], n_results: int = None,
                                batch_size: int = 5) -> List[Any]:
        all_results = []
        step = 0

        while True:
            heuristics = [next(request_synthesizer) for _ in range(step, step + batch_size)]

            if n_results is not None:
                heuristics = heuristics[:n_results]
                n_results -= batch_size

            _results = await self.request_pool(heuristics=heuristics)
            responses = [_result for _result in _results if not_stop_until(_result)]

            all_results.extend(responses)
            step += batch_size

            if len(responses) != batch_size:
                break

        return all_results

    async def multi_page_request(self, direct_url: str, offset: int = 0, results_per_page: int = 20,
                                 n_collect: int = None, list_key: str or list = None, **kwargs) -> List:
        if offset < 0:
            raise ValueError('Offset must be greater than or equal to zero.')

        elif n_collect and n_collect < 1:
            raise ValueError('n_collect must be greater than or equal to one.')

        def synthesizer():
            start_page = offset // results_per_page + 1

            while True:
                yield {'url': direct_url,
                       'params': {'page': start_page},
                       'output': 'json',
                       'method': 'get'} | kwargs

                start_page += 1

        n_results = (n_collect // results_per_page + int(n_collect % results_per_page != 0)) if n_collect else None

        data = await self.run_until_no_stop(request_synthesizer=synthesizer(),
                                            not_stop_until=lambda result: result != [],
                                            n_results=n_results)

        adjusted_start = offset % results_per_page

        if isinstance(list_key, list):
            records = [[record for page in data for record in page.get(key, [])]
                       for key in list_key]

            if n_collect:
                return [record[adjusted_start:adjusted_start + n_collect] for record in records]

            return [record[adjusted_start:] for record in records]

        elif list_key is None:
            records = [record for page in data for record in page]

        else:
            records = [record for page in data for record in page.get(list_key, [])]

        if n_collect:
            return records[adjusted_start:adjusted_start + n_collect]

        return records[adjusted_start:]

    async def close_session(self):
        """
        Closes the aiohttp session.
        """
        await self.session.close()



# This line of code will allow shorter imports
from drukarnia_api.author import Author
from drukarnia_api.article import Article
from drukarnia_api.tag import Tag
from drukarnia_api.search import Search

from drukarnia_api.drukarnia_base import DrukarniaAPIError, DrukarniaElementDataError

print("Слава Україні 🇺🇦")



import re
from datetime import datetime
from typing import Dict, Tuple, List
from warnings import warn
from aiohttp import ClientSession

from drukarnia_api.drukarnia_base import DrukarniaElement
from drukarnia_api.shortcuts import data2authors, data2articles, data2tags

from typing import TYPE_CHECKING

if TYPE_CHECKING:   # always False, used for type hints
    from drukarnia_api.article import Article
    from drukarnia_api.tag import Tag


class Author(DrukarniaElement):
    def __init__(self, username: str = None, author_id: str = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._update_data({'username': username, '_id': author_id})

    async def login(self, email: str, password: str) -> None:
        """
        Log in the author with the provided email and password.
        """

        # Make a POST request to log in the author
        headers, info = await self.request('post',
                                           '/api/users/login',
                                           data={"password": password, "email": email},
                                           output=['headers', 'json'])

        if self.author_id and (self.author_id != info['user']['_id']):
            raise ValueError('You are trying to log into an unrelated author.')

        elif self.username and (self.username != info['user']['username']):
            raise ValueError('You are trying to log into an unrelated author.')

        elif not (self.author_id or self.username):
            warn("We weren't able to identify any relationship between the current Author data and the Drukarnia "
                 "User you are trying to log into. It may cause unexpected and fatal errors. Please consider "
                 "initializing the Author class with the same username and _id as your Drukarnia User.")

        self._update_data(info['user'])

        headers = str(headers)
        token = re.search(r'refreshToken=(.*?);', headers).group(1)
        device_id = re.search(r'deviceId=(.*?);', headers).group(1)

        self._update_headers({'Cookie': f'deviceId={device_id}; token={token};'})

    @DrukarniaElement._control_attr('author_id')
    async def get_followers(self, create_authors: bool = True, offset: int = 0, results_per_page: int = 20,
                            n_collect: int = None, *args, **kwargs) -> Tuple['Author'] or Tuple[Dict]:
        """
        Get the followers of the author.
        """

        # Make a request to get the followers of the author
        followers = await self.multi_page_request(f'/api/relationships/{self.author_id}/followers',
                                                  offset, results_per_page, n_collect, *args, **kwargs)

        if create_authors:
            followers = await data2authors(followers, self.session)

        return followers

    @DrukarniaElement._control_attr('author_id')
    async def get_followings(self, create_authors: bool = True, offset: int = 0, results_per_page: int = 20,
                             n_collect: int = None, *args, **kwargs) -> Tuple['Author'] or Tuple[Dict]:
        """
        Get the followings of the author.
        """

        # Make a request to get the followings of the author
        followings = await self.multi_page_request(f'/api/relationships/{self.author_id}/following',
                                                   offset, results_per_page, n_collect, *args, **kwargs)

        if create_authors:
            followings = await data2authors(followings, self.session)

        return followings

    @DrukarniaElement._is_authenticated
    async def get_notifications(self, offset: int = 0, results_per_page: int = 20,
                                n_collect: int = None, *args, **kwargs) -> List[Dict]:
        """
        Get the notifications of the author.
        """
        return await self.multi_page_request('/api/notifications',
                                             offset, results_per_page, n_collect,
                                             *args, **kwargs)

    @DrukarniaElement._is_authenticated
    async def get_reads_history(self, create_articles: bool = True, offset: int = 0,
                                results_per_page: int = 20, n_collect: int = None,
                                *args, **kwargs) -> List[Dict] or List['Article']:
        """
        Get the reading history of the author.
        """
        articles = await self.multi_page_request('/api/stats/reads/history',
                                                 offset, results_per_page, n_collect,
                                                 *args, **kwargs)

        if create_articles:
            articles = await data2articles(articles, self.session)

        return articles

    @DrukarniaElement._is_authenticated
    async def get_sections(self, preview: bool = True, **kwargs) -> List[dict]:
        """
        Get the sections of the author's articles.
        """
        return await self.request('get', f'/api/articles/bookmarks/lists?preview={str(preview).lower()}',
                                  output='json', **kwargs)

    @DrukarniaElement._is_authenticated
    async def create_section(self, name: str, **kwargs) -> List[Dict]:
        """
        Create a new section for the author's articles.
        """
        section_id = await self.request('get',
                                        '/api/articles/bookmarks/lists',
                                        data={"name": name},
                                        output='read', **kwargs)

        return section_id.decode('utf-8')

    @DrukarniaElement._is_authenticated
    async def delete_section(self, section_id: str, **kwargs) -> None:
        """
        Delete a section for the author's articles.
        """
        await self.request('delete', f'/api/articles/bookmarks/lists/{section_id}', **kwargs)

    @DrukarniaElement._is_authenticated
    async def subscribe_author(self, author_id: str, unsubscribe: bool = False) -> None:
        """
        Subscribe or unsubscribe to/from an author.
        """

        if unsubscribe:
            await self.request('delete', f'/api/relationships/subscribe/{author_id}')
            return None

        await self.request('post', f'/api/relationships/subscribe/{author_id}')

    @DrukarniaElement._is_authenticated
    async def subscribe_tag(self, tag_id: str, unsubscribe: bool = False) -> None:
        """
        Subscribe or unsubscribe to/from a tag.
        """
        if unsubscribe:
            await self.request('delete', f'/api/preferences/tags/{tag_id}')
            return None

        await self.request('put', f'/api/preferences/tags/{tag_id}')

    @DrukarniaElement._is_authenticated
    async def block_tag(self, tag_id: str, unblock: bool = False) -> None:
        """
        Block or unblock an author.
        """
        if unblock:
            await self.request('put', f'/api/preferences/tags/{tag_id}/block', data={"isBlocked": False})
            return None

        await self.request('put', f'/api/preferences/tags/{tag_id}/block', data={"isBlocked": True})

    @DrukarniaElement._is_authenticated
    async def block_author(self, author_id: str, unblock: bool = False) -> None:
        """
        Block or unblock an author.
        """
        if unblock:
            await self.request('patch', f'/api/relationships/block/{author_id}')
            return None

        await self.request('put', f'/api/relationships/block/{author_id}')

    @DrukarniaElement._is_authenticated
    async def get_blocked(self, create_authors: bool = False) -> List[Dict] or Tuple['Author']:
        """
        Get the authors blocked by the current author.
        """

        authors = await self.request('get', '/api/relationships/blocked', output='json')

        if create_authors:
            authors = await data2authors(authors, self.session)

        return authors

    @DrukarniaElement._is_authenticated
    async def change_password(self, old_password: str, new_password: str, **kwargs) -> None:
        """
        Change the author's password.
        """
        await self.request('patch', f'/api/users/login/password',
                           data={"oldPassword": old_password, "newPassword": new_password},
                           output='read', **kwargs)

    @DrukarniaElement._is_authenticated
    async def change_user_info(self, name: str = None, description: str = None, username: str = None,
                               description_short: str = None, socials: dict = None, donate_url: str = None) -> str:
        """
        Change the author's user information.
        """
        info2patch = {"name": name, "description": description, "username": username,
                      "descriptionShort": description_short,
                      "socials": socials, "donateUrl": donate_url}

        info2patch = {key: value for key, value in info2patch.items() if value is not None}

        response = await self.request('patch', '/api/users', data=info2patch, output='read')
        return response.decode('utf-8')

    @DrukarniaElement._is_authenticated
    async def change_email(self, current_password: str, new_email: str, **kwargs) -> None:
        """
        Change the author's email.
        """
        await self.request('patch', f'/api/users/login/email',
                           data={"currentPassword": current_password, "newEmail": new_email},
                           **kwargs)

    @DrukarniaElement._control_attr('username')
    async def collect_data(self, return_: bool = False) -> Dict or None:
        """
        Collect the author's data and update the object's attributes.
        """

        data = await self.request('get', '/api/users/profile/{username}'.format(username=self.username),
                                  output='json')

        self._update_data(data)

        if return_:
            return data

    @property
    def username(self) -> str:
        return self._get_basetype_from_data('username', str)

    @property
    def avatar(self) -> str:
        return self._get_basetype_from_data('avatar', str)

    @property
    def donate_url(self) -> str:
        return self._get_basetype_from_data('donateUrl', str)

    @property
    def socials(self) -> Dict:
        return self._get_basetype_from_data('socials', dict)

    @property
    def author_id(self) -> str:
        return self._get_basetype_from_data('_id', str)

    @property
    def name(self) -> str:
        return self._get_basetype_from_data('name', str)

    @property
    def description(self) -> str:
        return self._get_basetype_from_data('description', str)

    @property
    def description_short(self) -> str:
        return self._get_basetype_from_data('descriptionShort', str)

    @property
    def created_at(self) -> datetime:
        return self._get_datetime_from_author_data('createdAt')

    @property
    def following_num(self) -> int:
        return self._get_basetype_from_data('followingNum', int)

    @property
    def followers_num(self) -> int:
        return self._get_basetype_from_data('followersNum', int)

    @property
    def read_num(self) -> int:
        return self._get_basetype_from_data('readNum', int)

    @property
    async def articles(self) -> Tuple['Article']:
        return await data2articles(self._access_data('articles', []), self.session)

    @property
    async def author_tags(self) -> Tuple['Tag']:
        return await data2tags(self._access_data('authorTags', []), self.session)

    @property
    def relationships(self) -> Dict:
        return self._get_basetype_from_data('relationships', dict)

    @staticmethod
    async def from_records(session: ClientSession, new_data: dict) -> 'Author':
        """
        Create an Author instance from records.
        """
        new_author = Author(session=session)
        new_author._update_data(new_data)
        return new_author

    def __hash__(self) -> int:
        return hash(self.author_id or self.username)



from typing import Tuple
from aiohttp import ClientSession
import asyncio


async def data2tags(tags_data: list or None, session: ClientSession) -> Tuple:
    """
    Converts a list of tag data into Tag objects asynchronously.

    Args:
        tags_data (list or None): List of tag data.
        session (ClientSession): aiohttp ClientSession object.

    Returns:
        Tuple: Tuple of Tag objects.
    """
    if not tags_data:
        return ()

    from drukarnia_api.tag import Tag

    tasks = [Tag.from_records(session, tag) for tag in tags_data]

    return await asyncio.gather(*tasks)


async def data2authors(authors_data: list or None, session: ClientSession) -> Tuple:
    """
    Converts a list of author data into Author objects asynchronously.

    Args:
        authors_data (list or None): List of author data.
        session (ClientSession): aiohttp ClientSession object.

    Returns:
        Tuple: Tuple of Author objects.
    """
    if not authors_data:
        return ()

    from drukarnia_api.author import Author

    tasks = [Author.from_records(session, author) for author in authors_data]

    return await asyncio.gather(*tasks)


async def data2articles(articles_data: list or None, session: ClientSession) -> Tuple:
    """
    Converts a list of article data into Article objects asynchronously.

    Args:
        articles_data (list or None): List of article data.
        session (ClientSession): aiohttp ClientSession object.

    Returns:
        Tuple: Tuple of Article objects.
    """
    if not articles_data:
        return ()

    from drukarnia_api.article import Article

    tasks = [Article.from_records(session, article) for article in articles_data]

    return await asyncio.gather(*tasks)



from drukarnia_api.drukarnia_base import Connection
from drukarnia_api.shortcuts import data2authors, data2articles, data2tags

from typing import TYPE_CHECKING, Tuple, Dict

if TYPE_CHECKING:   # always False, used for type hints
    from drukarnia_api.article import Article
    from drukarnia_api.tag import Tag
    from drukarnia_api.author import Author


class Search(Connection):
    async def find_author(self, query: str, create_authors: bool = True, with_relations: bool = False,
                          offset: int = 0, results_per_page: int = 20, n_collect: int = None,
                          *args, **kwargs) -> Tuple['Author'] or Tuple[Dict]:
        """
        Search for authors.
        """

        with_relations = str(with_relations).lower()

        # Make a request to get authors
        authors = await self.multi_page_request(f'/api/users/info?name={query}&withRelationships={with_relations}',
                                                offset, results_per_page, n_collect, *args, **kwargs)

        if create_authors:
            authors = await data2authors(authors, self.session)

        return authors

    async def find_articles(self, query: str, create_articles: bool = True,
                            offset: int = 0, results_per_page: int = 20, n_collect: int = None,
                            *args, **kwargs) -> Tuple['Article'] or Tuple[Dict]:
        """
        Search for articles.
        """

        # Make a request to get articles
        articles = await self.multi_page_request(f'/api/articles/search?name={query}',
                                                 offset, results_per_page, n_collect, *args, **kwargs)

        if create_articles:
            articles = await data2articles(articles, self.session)

        return articles

    async def find_tags(self, query: str, create_tags: bool = True, get_articles: bool = False,
                        create_articles: bool = True, offset: int = 0, results_per_page: int = 20,
                        n_collect: int = None, *args, **kwargs) -> Tuple['Tag'] or Tuple[Dict]:
        """
        Search for tags.
        """

        # Make a request to get articles
        tags, articles = await self.multi_page_request(f'/api/articles/search/tags?text={query}',
                                                       list_key=['tags', 'articles'],
                                                       offset=offset,
                                                       results_per_page=results_per_page,
                                                       n_collect=n_collect, *args, **kwargs)

        if get_articles:
            if create_articles:
                articles = await data2articles(articles, self.session)

            return tags, articles

        if create_tags:
            tags = await data2tags(tags, self.session)

        return tags



from aiohttp import ClientSession

from drukarnia_api.drukarnia_base import DrukarniaElement
from drukarnia_api.shortcuts import data2articles, data2tags, data2authors

from typing import TYPE_CHECKING, Tuple, Dict

if TYPE_CHECKING:   # always False, used for type hints
    from drukarnia_api.article import Article
    from drukarnia_api.author import Author


class Tag(DrukarniaElement):
    def __init__(self, slug_name: str = None, tag_id: str = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Update the data with slug_name and tag_id
        self._update_data({'slug': slug_name, '_id': tag_id})

    @DrukarniaElement._control_attr('slug')
    async def get_articles(self, create_articles: bool = True,
                           offset: int = 0, results_per_page: int = 20, n_collect: int = None,
                           **kwargs) -> Tuple['Article'] or Tuple[Dict]:
        """
        Get the followers of the author.
        """

        # Make a request to get the followers of the author
        articles = await self.multi_page_request(f'/api/articles/tags/{self.slug}',
                                                 offset, results_per_page, n_collect, list_key='articles',
                                                 **kwargs)
        if create_articles:
            articles = await data2articles(articles, self.session)

        return articles

    @DrukarniaElement._control_attr('tag_id')
    async def related_tags(self, create_tags: bool = True) -> Tuple['Tag'] or Tuple[Dict]:
        """
        Get the followers of the author.
        """

        # Make a request to get the followers of the author
        tags = await self.request('get', f'/api/articles/tags/{self.tag_id}/related?page=1', output='json')
        if create_tags:
            tags = await data2tags(tags, self.session)

        return tags

    @DrukarniaElement._control_attr('tag_id')
    async def related_authors(self, create_tags: bool = True,) -> Tuple['Author'] or Tuple[Dict]:
        """
        Get the followers of the author.
        """

        # Make a request to get the followers of the author
        authors = await self.request('get', f'/api/users/tags/{self.tag_id}/related', output='json')
        if create_tags:
            authors = await data2authors(authors, self.session)

        return authors

    @DrukarniaElement._control_attr('slug')
    async def collect_data(self, return_: bool = False):
        result = await self.request('get', f'/api/articles/tags/{self.slug}?page=1', output='json')

        if result.get('articles', None):
            del result['articles']

        self._update_data(result)

        if return_:
            return result

    @property
    def slug(self):
        """
        Get the slug property of the Tag.
        """
        return self._access_data('slug', str)

    @property
    def created_at(self):
        """
        Get the created_at property of the Tag.
        """
        return self._get_datetime_from_author_data('createdAt')

    @property
    def default(self):
        """
        Get the default property of the Tag.
        """
        return self._get_basetype_from_data('default', bool)

    @property
    def mentions_num(self):
        """
        Get the mentions_num property of the Tag.
        """
        return self._get_basetype_from_data('mentionsNum', int)

    @property
    def name(self):
        """
        Get the name property of the Tag.
        """
        return self._access_data('name', None)

    @property
    def tag_id(self):
        """
        Get the _id property of the Tag.
        """
        return self._access_data('_id', None)

    @property
    def relationships(self):
        return self._access_data('relationships', None)

    @staticmethod
    async def from_records(session: ClientSession, new_data: dict) -> 'Tag':
        """
        Create a new Tag instance from records.

        Args:
            session (ClientSession): A session object for making HTTP requests.
            new_data (dict): Data to update the new Tag instance with.

        Returns:
            Tag: A new instance of the Tag class.
        """
        new_tag = Tag(session=session)
        new_tag._update_data(new_data)

        return new_tag
